/*

Copyright (C) University of Oxford, 2005-2009

University of Oxford means the Chancellor, Masters and Scholars of the
University of Oxford, having an administrative office at Wellington
Square, Oxford OX1 2JD, UK.

This file is part of Chaste.

Chaste is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published
by the Free Software Foundation, either version 2.1 of the License, or
(at your option) any later version.

Chaste is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details. The offer of Chaste under the terms of the
License is subject to the License being interpreted in accordance with
English Law and subject to any action against the University of Oxford
being under the jurisdiction of the English Courts.

You should have received a copy of the GNU Lesser General Public License
along with Chaste. If not, see <http://www.gnu.org/licenses/>.

*/


#ifndef POLEZEROMATERIALLAW_HPP_
#define POLEZEROMATERIALLAW_HPP_


#include "AbstractIncompressibleMaterialLaw.hpp"
#include "Exception.hpp"



/**
 *  Pole-zero material law
 *
 *  NOTE: CHANGED THE CODE AT THE MOMENT TO NOT USE THE POSITIVE PART BIT BELOW
 *  AS THEN IT IS NOT TWICE DIFFERENTIABLE
 *
 *  W = \Sum_{M,N=1..3} k_{MN} ([ E_{MN} ]_+)^2 / (a_{MN} - E_{MN})^b_{MN}
 *
 *  Note the positive part operator in the numerator, so that
 *  the term in W corresponding to M,N is zero if E_MN < 0. (This differs
 *  from the original pole-zero paper but seems to be what they meant..)
 *
 *  Note that is the parameters k4,k5,k6,a4,a5,a6 etc are known, then
 *  k01=k10=0.5*k4 and similarly with k5,k6, but a01=a10=a4 etc.
 *
 *  Not isotropic, so inherits directly from AbstractIncompressibleMaterialLaw
 */
template<unsigned DIM>
class PoleZeroMaterialLaw : public AbstractIncompressibleMaterialLaw<DIM>
{
friend class TestMaterialLaws;

private :
    std::vector<std::vector<double> > mK;
    std::vector<std::vector<double> > mA;
    std::vector<std::vector<double> > mB;

    c_matrix<double,DIM,DIM> mIdentity;

protected :
    /**
     *  Protected default constructor doing nothing. Just saw inherited classes
     *  can be instantiated and THEN set up the parameters
     */
    PoleZeroMaterialLaw();

    /**
     *  Set k, a, and b. To be called by the constuctor or a child class
     *  Set comments for constructor.
     */
    void SetParameters(std::vector<std::vector<double> > k,
                       std::vector<std::vector<double> > a,
                       std::vector<std::vector<double> > b);
  

public :
    /**
     *  Constructor, taking in parameters k_i, a_i, b_i as matrices.
     *  These matrices must be of size DIM-by-DIM and must be symmetric
     *
     *  Note: using the k_1..k_6 convention,  k_4 = 2*k[0][1] = 2*k[1][0], etc
     */
     PoleZeroMaterialLaw(std::vector<std::vector<double> > k,
                         std::vector<std::vector<double> > a,
                         std::vector<std::vector<double> > b);

    void ComputeStressAndStressDerivative(c_matrix<double,DIM,DIM>& C,
                                          c_matrix<double,DIM,DIM>& invC,
                                          double                    pressure,
                                          c_matrix<double,DIM,DIM>& T,
                                          FourthOrderTensor2<DIM>&  dTdE,
                                          bool                      computeDTdE);

    double GetZeroStrainPressure();

    /** Scale the dimensional material parameters (ie the K's) */
    void ScaleMaterialParameters(double scaleFactor);
};

#endif /*POLEZEROMATERIALLAW_HPP_*/
